/*
 * ?? submission "Vanadium"
 * test SMMU determinism
 * Copyright (C) Jack Whitham 2010
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "tetra.h"
#include "smmu.h"
#include "cache.h"
#include "memlib.h"

#include "u64.hh"

#define NUM_TESTS       1000
#define NUM_SIZES       9
#define NUM_LATENCIES   3
#define SMALL_MEM_SIZE  (1 << 16)

static ulongint Test_New (Tetra_Params * tp, 
            ulongint previous, ulongint max_size)
{
    if (tp->force_worst) {
        return New(tp, previous, max_size);
    } else {
        ulongint out;
        ulongint mask = previous & PTR_ID_MASK;

        long int out1 = 0;
        lrand48_r(&tp->rng_state, &out1);
        assert (out1 >= 0);
        out1 %= SMALL_MEM_SIZE;
        out = (ulongint) out1;
        out >>= tp->align_bits;
        out <<= tp->align_bits;
        out += USE_MEM_START;
        return out | mask;
    }
}

int main (void)
{
    ulongint        i, j, k;

    for (i = 0; tetra_methods[ i ] != NULL; i++) {
        Tetra_Params *  tp;
        const char *    name = tetra_descs[ i ];
        ulongint        best, worst, this_time;

        assert (name != NULL);
        tp = Init_Params ();
        tp->new_fn = Test_New;

        for (k = 0; k < 2; k ++) {
            printf("%s (%d): ", name, (int) k);
            fflush(stdout);

            tp->smmu_very_deterministic = k;
            tp->clock = 0;
            tp->force_worst = 1;
            Init_SMMU(tp, 16, CM_OPTION_B);
            tetra_methods[ i ](tp);
            tp->flush_fn(tp->mem_model_cb);
            best = worst = tp->clock;
            printf("%"FU" ", worst);
            fflush(stdout);

            tp->force_worst = 0;
            for (j = 0; j < NUM_TESTS; j ++) {
                tp->clock = 0;
                tetra_methods[ i ](tp);
                tp->flush_fn(tp->mem_model_cb);
                this_time = tp->clock;
                if (this_time > worst) {
                    printf("\nOh no! This time is %"FU", worse than %"FU"\n",
                        this_time, worst);
                    exit(1);
                }
                if (this_time < best) {
                    best = this_time;
                }
            }
            tp->destroy_fn(tp->mem_model_cb);
            printf("%"FU"", best);
            if (tp->smmu_very_deterministic) {
                if (best != worst) {
                    printf("\nOh no! best != worst in deterministic mode.\n");
                    exit(1);
                } else {
                    printf(" :)");
                }
            } else {
                printf(" var %1.3e", 
                    (((double) worst) / ((double) best)) - 1.0);
            }
            printf("\n");
        }
        free (tp);
    }
    printf("All tests passed.\n");
    return 0;
}


