/*
 * ECRTS 2010 submission "Sulphur"
 * library functions NEW(), RANDOM()
 * Copyright (C) Jack Whitham 2010
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "tetra.h"
#include "u64.hh"

ulongint New (Tetra_Params * tp, ulongint previous, ulongint max_size)
{
    ulongint out;
    ulongint mask = previous & PTR_ID_MASK;

    if (tp->force_worst) {
        out = tp->new_position;
        assert ((out + max_size) < USE_MEM_SIZE);
        tp->new_position = out + max_size + MIN_SKIP;
    } else {
        long int out1 = 0;
        lrand48_r(&tp->rng_state, &out1);
        assert (out1 >= 0);
        out1 %= USE_MEM_SIZE;
        out = (ulongint) out1;
    }
    out >>= tp->align_bits;
    out <<= tp->align_bits;
    out += USE_MEM_START;
    return out | mask;
}

ulongint Random (Tetra_Params * tp, ulongint bound)
{
    long int out1 = 0;
    lrand48_r(&tp->rng_state, &out1);
    assert (out1 >= 0);
    out1 %= bound;
    return (ulongint) out1;
}

ulongint Calc_Txfer (Tetra_Params * tp, 
                    ulongint size, ulongint min_cycles)
{
    ulongint        transactions = 0;
    ulongint        followers = 0;
    ulongint        est_cost = 0;

    assert (size > 0);
    assert (size < MEM_SIZE);
    if (size == 1) {
        est_cost = tp->buslat + 1;
    } else {
        ulongint s = size % tp->max_transaction_size;
        if (size >= tp->max_transaction_size) {
            transactions += size / tp->max_transaction_size;
            followers += (((tp->max_transaction_size - 1) / tp->bus_width)
                                * (size / tp->max_transaction_size));
        }
        if (s > 0) {
            transactions ++;
            followers += (s - 1) / tp->bus_width;
        }

        est_cost = (transactions * tp->buslat) + followers;
    }

    if (est_cost < min_cycles) {
        est_cost = min_cycles;
    }
    return est_cost;
}

static void Do_Txfer (Tetra_Params * tp, 
                    ulongint size, ulongint min_cycles)
{
    tp->clock += Calc_Txfer(tp, size, min_cycles);
}

static void Do_Tick (Tetra_Params * tp, Tick_Type tt)
{
    if (tt == TICK_ON_CHIP_ACCESS) {
        tp->clock ++;
    }
}

Tetra_Params * Init_Params ( void )
{
    Tetra_Params *  tp = calloc(1, sizeof(Tetra_Params));

    tp->ro_max = 3;
    tp->loop_size = 20000;
    tp->page_size = 256;
    tp->tile_size = 128;
    tp->tiling_flag = 0;
    tp->bytes_on_chip = 16384;
    tp->buslat = 49;
    tp->max_transaction_size = 64;
    tp->bus_width = 4;
    tp->new_fn = New;
    tp->random_fn = Random;
    tp->calc_txfer_fn = Calc_Txfer;
    tp->do_txfer_fn = Do_Txfer;
    tp->do_tick_fn = Do_Tick;
    tp->force_worst = 0;
    tp->align_bits = 6;
    srand48_r(1, &tp->rng_state);

    return tp;
}

