/*
 * ECRTS 2010 submission "Sulphur"
 * Internal interface definition
 * Copyright (C) Jack Whitham 2010
 */

#ifndef TETRA_H
#define TETRA_H

#include <stdint.h>
#include <stdlib.h>


#ifndef ulongint
#define ulongint uint64_t
#endif

#define PTR_ID_OFFSET   48
#define PTR_ID_MASK     (((ulongint) 0xff) << PTR_ID_OFFSET )
#define ADDRESS_MASK    (((((ulongint) 1) << PTR_ID_OFFSET ) >> 8 ) - 1 )
#define MEM_SIZE        (( ADDRESS_MASK >> 4 ) + 1 )
#define MIN_SKIP        256
#define USE_MEM_START   (( ADDRESS_MASK >> 6 ) + 1 )
#define USE_MEM_SIZE    (( ADDRESS_MASK >> 6 ) + 1 )

typedef enum {
    TICK_LOOP_OVERHEAD, TICK_ON_CHIP_ACCESS
    } Tick_Type;

typedef struct Tetra_Params_struct {
    /* Set by memlab. */
    ulongint        ro_max, loop_size;
    ulongint        page_size, tile_size;
    ulongint        (* new_fn) (struct Tetra_Params_struct * tp, 
                        ulongint previous, ulongint max_size);
    ulongint        (* random_fn) (struct Tetra_Params_struct * tp, 
                        ulongint bound);
    ulongint        (* calc_txfer_fn) (struct Tetra_Params_struct * tp, 
                        ulongint size, ulongint min_cycles);
    void            (* do_txfer_fn) (struct Tetra_Params_struct * tp, 
                        ulongint size, ulongint min_cycles);
    void            (* do_tick_fn) (struct Tetra_Params_struct * tp, 
                        Tick_Type tt);
    ulongint        buslat, align_bits, max_transaction_size, bus_width;
    ulongint        bytes_on_chip;
    struct drand48_data rng_state;
    ulongint        new_position;
    int             tiling_flag;
    int             force_worst;
    int             smmu_powers_of_two;
    int             smmu_very_deterministic;

    /* Initialised to zero by memlab.
     * Set by memory component model. */
    int             page_flag;
    ulongint        (* close_fn) (void * cb, ulongint value);
    ulongint        (* open_fn) (void * cb, ulongint value,
                        ulongint size, ulongint * offset, int read_only);
    void            (* read_fn) (void * cb, ulongint value, 
                        const char * comment);
    void            (* write_fn) (void * cb, ulongint value, 
                        const char * comment);
    int             (* worth_copying_fn) (void * cb, ulongint size, 
                        ulongint access_count, int read_only);
    void            (* destroy_fn) (void * cb);
    void            (* flush_fn) (void * cb);
    void *          mem_model_cb;

    /* Tracks events */
    ulongint        clock;
    void *          user_cb;
} Tetra_Params ;

typedef void (* Tetra_Method) (Tetra_Params * tp);

extern Tetra_Method tetra_methods [];
extern const char * tetra_descs [];

#endif

